# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for recipe_manager.py."""

import copy

from absl.testing import absltest
from absl.testing import parameterized

from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import algorithm_manager_api
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe_manager

_ComputePrecision = qtyping.ComputePrecision
_TFLOpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_TensorDataType = qtyping.TensorDataType
_AlgorithmName = recipe_manager.AlgorithmName
_QuantGranularity = qtyping.QuantGranularity


def _sample_check_op_config_func(op_name, op_config, _):
  if (
      op_config.weight_tensor_config is not None
      and op_config.weight_tensor_config.num_bits == 17
  ):
    raise ValueError(f'Unsupported number of bits for op: {op_name}.')


def _add_default_int8xint8_integer_recipe(recipe_manager_object):
  recipe_manager_object.add_quantization_config(
      regex='.*',
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
      op_config=qtyping.OpQuantizationConfig(
          activation_tensor_config=_TensorQuantConfig(
              num_bits=8, symmetric=False
          ),
          weight_tensor_config=_TensorQuantConfig(num_bits=8),
          compute_precision=_ComputePrecision.INTEGER,  # SRQ.
      ),
  )


# register some currently unsupported ops for testing purposes
def _register_testing_op(algorithm_key, tfl_op):
  # Sample functions for test cases.
  def _sample_init_qsvs(*_, **__):
    return {'name': dict()}

  def _sample_calibration_func(*_, **__):
    return {'name2': dict()}

  def _sample_materialize_func(*_, **__):
    return []

  algorithm_manager.register_op_quant_config_validation_func(
      algorithm_key, _sample_check_op_config_func
  )
  algorithm_manager.register_quantized_op(
      algorithm_key,
      tfl_op,
      _sample_init_qsvs,
      _sample_calibration_func,
      _sample_materialize_func,
  )


class ConfiguratorTest(parameterized.TestCase):
  """Test cases for the flax quantizer Configurator."""

  def setUp(self):
    super().setUp()
    self._setup_algorithm_manager()
    self._recipe_manager = recipe_manager.RecipeManager()
    self._testing_ops = [
        _TFLOpName.BATCH_MATMUL,
        _TFLOpName.FULLY_CONNECTED,
        _TFLOpName.DEPTHWISE_CONV_2D,
    ]
    for op in self._testing_ops:
      _register_testing_op(_AlgorithmName.MIN_MAX_UNIFORM_QUANT, op)
      _register_testing_op('GPTQ', op)

  def tearDown(self):
    self._restore_algorithm_manager()
    super().tearDown()

  def _setup_algorithm_manager(self):
    self._original_alg_manager_instance = (
        algorithm_manager._alg_manager_instance
    )
    self._new_alg_manager_instance = algorithm_manager_api.AlgorithmManagerApi()

    # Deep copy the state from original instance
    self._new_alg_manager_instance._config_check_registry = (
        self._original_alg_manager_instance._config_check_registry.copy()
    )
    self._new_alg_manager_instance._config_check_policy_registry = (
        self._original_alg_manager_instance._config_check_policy_registry.copy()
    )
    self._new_alg_manager_instance._algorithm_registry = {}
    for k, v in self._original_alg_manager_instance._algorithm_registry.items():
      # v is QuantizationAlgorithmInfo
      new_v = copy.copy(v)
      new_v.quantized_ops = v.quantized_ops.copy()
      self._new_alg_manager_instance._algorithm_registry[k] = new_v

    algorithm_manager._alg_manager_instance = self._new_alg_manager_instance
    self._original_funcs = {}
    for func_name in [
        'get_quantization_func',
        'get_supported_ops',
        'get_init_qsv_func',
        'register_op_quant_config_validation_func',
        'register_config_check_policy_func',
        'register_quantized_op',
        'is_op_registered',
        'is_algorithm_registered',
        'check_op_quantization_config',
    ]:
      self._original_funcs[func_name] = getattr(algorithm_manager, func_name)
      if hasattr(self._new_alg_manager_instance, func_name):
        setattr(
            algorithm_manager,
            func_name,
            getattr(self._new_alg_manager_instance, func_name),
        )
      elif func_name == 'register_config_check_policy_func':
        setattr(
            algorithm_manager,
            func_name,
            self._new_alg_manager_instance.register_config_check_policy,
        )

  def _restore_algorithm_manager(self):
    algorithm_manager._alg_manager_instance = (
        self._original_alg_manager_instance
    )
    for func_name, func in self._original_funcs.items():
      setattr(algorithm_manager, func_name, func)

  def test_add_get_quantization_config(self):
    # Int8 DRQ all ops under "Dense".
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        ),
    )

    # Int8 weight-only FullyConnected configuration under "Dense_3".
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_3/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )
    # Int4 DRQ BatchMatmul configuration under "Dense_3".
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_3/.*',
        operation_name=_TFLOpName.BATCH_MATMUL,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=4),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        ),
    )

    # Return NO_QUANT if not match.
    alg_key, _ = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense_1/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.NO_QUANTIZE)
    alg_key, _ = self._recipe_manager.get_quantization_configs(
        _TFLOpName.DEPTHWISE_CONV_2D, 'model/Dense_3/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.NO_QUANTIZE)

    # Check _TFLOperationKey.ALL
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.DEPTHWISE_CONV_2D, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    # DRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)

    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.BATCH_MATMUL, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    # DRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)

    # Check conflicts handling.
    # Int8 Weight-only for FC under "Dense", this should only overwrite FC but
    # leave others unchanged.
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
        ),
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    # WEIGHT_ONLY check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.FLOAT)
    alg_key, _ = self._recipe_manager.get_quantization_configs(
        _TFLOpName.BATCH_MATMUL, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)

    # Reset all ops, this time with 4 bits DRQ.
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=4),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        ),
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    # DRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)
    self.assertEqual(weight_tensor_config.num_bits, 4)
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.BATCH_MATMUL, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    # DRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)
    self.assertEqual(weight_tensor_config.num_bits, 4)

    # Overwrite all FC.
    self._recipe_manager.add_quantization_config(
        regex='.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=3),
        ),
    )
    # FC config is overridden.
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense_3/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 3)
    # No overridden for batch matmul.
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.BATCH_MATMUL, 'model/Dense_3/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 4)

  def test_add_unsupported_quantization_config(self):
    error_message = 'Unsupported operation'
    # Add unregistered operations.
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._recipe_manager.add_quantization_config(
          regex='.*/Dense/.*',
          operation_name=_TFLOpName.CUSTOM_OP,
          algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
          op_config=qtyping.OpQuantizationConfig(
              weight_tensor_config=_TensorQuantConfig(num_bits=8),
              compute_precision=_ComputePrecision.INTEGER,  # DRQ.
          ),
      )

  def test_add_unsupported_num_bits_raise_error(self):
    test_op_name = _TFLOpName.FULLY_CONNECTED
    error_message = f'Unsupported number of bits for op: {test_op_name}.'
    # Add unregistered operation
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._recipe_manager.add_quantization_config(
          regex='.*/Dense/.*',
          operation_name=test_op_name,
          algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
          op_config=qtyping.OpQuantizationConfig(
              weight_tensor_config=_TensorQuantConfig(num_bits=17),
          ),
      )

  def test_add_unsupported_skip_successful(self):
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_3/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=17),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
            skip_checks=True,
        ),
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense_3/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertIsNone(op_config.activation_tensor_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 17)
    # DRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)

  def test_add_unsupported_algorithm_key_raise_error(self):
    error_message = 'Unsupported algorithm key'
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._recipe_manager.add_quantization_config(
          regex='.*/Dense/.*',
          operation_name=_TFLOpName.FULLY_CONNECTED,
          algorithm_key='decomposed_hadamard',
          op_config=qtyping.OpQuantizationConfig(
              weight_tensor_config=_TensorQuantConfig(num_bits=8),
          ),
      )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._recipe_manager.add_quantization_config(
          regex='.*/Dense/.*',
          operation_name=_TFLOpName.ALL_SUPPORTED,
          algorithm_key='decomposed_hadamard',
          op_config=qtyping.OpQuantizationConfig(
              weight_tensor_config=_TensorQuantConfig(num_bits=8),
          ),
      )

  def test_add_dynamic_config(self):
    self._recipe_manager.add_dynamic_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        num_bits=8,
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)
    self.assertFalse(op_config.explicit_dequantize)
    self.assertIsNone(op_config.activation_tensor_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 8)
    self.assertTrue(weight_tensor_config.symmetric)
    self.assertEqual(
        weight_tensor_config.granularity,
        _QuantGranularity.CHANNELWISE,
    )

  @parameterized.parameters(4, 8)
  def test_add_weight_only_config_int(self, num_bits):
    self._recipe_manager.add_weight_only_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        num_bits=num_bits,
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertEqual(op_config.compute_precision, _ComputePrecision.FLOAT)
    self.assertTrue(op_config.explicit_dequantize)
    self.assertIsNone(op_config.activation_tensor_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, num_bits)
    self.assertTrue(weight_tensor_config.symmetric)
    self.assertEqual(
        weight_tensor_config.granularity,
        _QuantGranularity.CHANNELWISE,
    )
    self.assertEqual(weight_tensor_config.dtype, _TensorDataType.INT)

  def test_add_weight_only_config_fp16(self):
    self._recipe_manager.add_weight_only_config(
        regex='.*/Dense2/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        num_bits=16,
        algorithm_key=_AlgorithmName.FLOAT_CASTING,
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense2/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.FLOAT_CASTING)
    self.assertEqual(op_config.compute_precision, _ComputePrecision.FLOAT)
    self.assertTrue(op_config.explicit_dequantize)
    self.assertIsNone(op_config.activation_tensor_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 16)
    self.assertTrue(weight_tensor_config.symmetric)
    self.assertEqual(
        weight_tensor_config.granularity,
        _QuantGranularity.CHANNELWISE,
    )
    self.assertEqual(weight_tensor_config.dtype, _TensorDataType.FLOAT)

  def test_add_weight_only_config_fp8_raise_error(self):
    error_message = (
        'float casting quantization config requires number of bits to be set'
        ' as 16'
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._recipe_manager.add_weight_only_config(
          regex='.*/Dense2/.*',
          operation_name=_TFLOpName.FULLY_CONNECTED,
          num_bits=8,
          algorithm_key=_AlgorithmName.FLOAT_CASTING,
      )

  def test_add_static_config(self):
    self._recipe_manager.add_static_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        activation_num_bits=8,
        weight_num_bits=4,
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)
    self.assertFalse(op_config.explicit_dequantize)
    activation_tensor_config = op_config.activation_tensor_config
    self.assertIsNotNone(activation_tensor_config)
    self.assertEqual(activation_tensor_config.num_bits, 8)
    self.assertFalse(activation_tensor_config.symmetric)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 4)
    self.assertTrue(weight_tensor_config.symmetric)
    self.assertEqual(
        weight_tensor_config.granularity,
        _QuantGranularity.CHANNELWISE,
    )

  def test_set_full_integer_quantization_config(self):
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    # Full integer setting is global
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense_3/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    op_act_config = op_config.activation_tensor_config
    self.assertIsNotNone(op_act_config)
    self.assertEqual(op_act_config.num_bits, 8)
    self.assertEqual(op_act_config.symmetric, False)
    self.assertEqual(
        op_act_config.granularity,
        _QuantGranularity.TENSORWISE,
    )
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 8)
    self.assertEqual(weight_tensor_config.symmetric, True)
    self.assertEqual(
        weight_tensor_config.granularity,
        _QuantGranularity.TENSORWISE,
    )

    # Change weight settings for Dense_3 FC
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_3/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=3),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        ),
    )
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense_3/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertIsNone(op_config.activation_tensor_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 3)
    # WEIGHT_ONLY check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)

    # Change the global setting to  int16
    self._recipe_manager.add_quantization_config(
        regex='.*',
        operation_name=_TFLOpName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            activation_tensor_config=_TensorQuantConfig(
                num_bits=16, symmetric=True
            ),
            weight_tensor_config=_TensorQuantConfig(num_bits=8, symmetric=True),
            compute_precision=_ComputePrecision.INTEGER,  # SRQ.
        ),
    )
    # This does not impact the special dense_3 case
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense_3/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertIsNone(op_config.activation_tensor_config)
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(weight_tensor_config.num_bits, 3)
    # WEIGHT_ONLY check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)

    # All the others will be int16
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.CONV_2D, 'model/Dense_31/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    op_act_config = op_config.activation_tensor_config
    self.assertIsNotNone(op_act_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    self.assertEqual(op_act_config.num_bits, 16)
    self.assertEqual(op_act_config.symmetric, True)
    self.assertEqual(
        op_act_config.granularity,
        _QuantGranularity.TENSORWISE,
    )
    self.assertEqual(weight_tensor_config.num_bits, 8)
    self.assertEqual(weight_tensor_config.symmetric, True)
    self.assertEqual(
        weight_tensor_config.granularity,
        _QuantGranularity.TENSORWISE,
    )

  def test_get_full_quantization_config(self):
    # Int8 asymetric full integer model.
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    # Default all BMM.
    self._recipe_manager.add_quantization_config(
        regex='.*',
        operation_name=_TFLOpName.BATCH_MATMUL,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    # Int8 DRQ FULLY_CONNECTED ops under "Dense".
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        ),
    )

    # Overwrite DRQ ALL ops under "Dense".
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=4),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    # Overwrite "Dense_1" to only quantize FullyConnected.
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_1/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=6),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    # Add BMM to "Dense_1".
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_1/.*',
        operation_name=_TFLOpName.BATCH_MATMUL,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=3),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    expected_full_quantization_config = [
        {
            'regex': '.*',
            'operation': '*',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'activation_tensor_config': {
                    'num_bits': 8,
                    'symmetric': False,
                    'granularity': _QuantGranularity.TENSORWISE,
                    'dtype': 'INT',
                },
                'weight_tensor_config': {
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': _QuantGranularity.TENSORWISE,
                    'dtype': 'INT',
                },
                # WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.INTEGER,
                'explicit_dequantize': False,
                'skip_checks': False,
                'min_weight_elements': 0,
            },
        },
        {
            'regex': '.*',
            'operation': 'BATCH_MATMUL',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': _QuantGranularity.TENSORWISE,
                },
                # WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
                'skip_checks': False,
                'min_weight_elements': 0,
            },
        },
        {
            'regex': '.*/Dense/.*',
            'operation': '*',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 4,
                    'symmetric': True,
                    'granularity': _QuantGranularity.TENSORWISE,
                },
                # WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
                'skip_checks': False,
                'min_weight_elements': 0,
            },
        },
        {
            'regex': '.*/Dense_1/.*',
            'operation': 'FULLY_CONNECTED',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 6,
                    'symmetric': True,
                    'granularity': _QuantGranularity.TENSORWISE,
                },
                # WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
                'skip_checks': False,
                'min_weight_elements': 0,
            },
        },
        {
            'regex': '.*/Dense_1/.*',
            'operation': 'BATCH_MATMUL',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 3,
                    'symmetric': True,
                    'granularity': _QuantGranularity.TENSORWISE,
                },
                # WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
                'skip_checks': False,
                'min_weight_elements': 0,
            },
        },
    ]
    self.assertEqual(
        expected_full_quantization_config,
        self._recipe_manager.get_quantization_recipe(),
    )

  def test_get_quantization_configs_with_no_quantize_overwrite(self):
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
        ),
    )
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.NO_QUANTIZE,
    )

    # Fully connected will be overwritten to no quantization.
    alg_key, _ = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.NO_QUANTIZE)
    # Other ops will be quantized.
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.CONV_2D, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    # WEIGHT_ONLY check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.FLOAT)
    self.assertEqual(weight_tensor_config.num_bits, 8)

  def test_load_from_full_quantization_config(self):
    full_quantization_config = [
        {
            'regex': '.*',
            'operation': 'BATCH_MATMUL',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': _QuantGranularity.CHANNELWISE,
                },
                # WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': False,
            },
        },
        {
            'regex': '.*/Dense/.*',
            'operation': '*',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 4,
                    'symmetric': False,
                    'granularity': _QuantGranularity.CHANNELWISE,
                },
                # DRQ.
                'compute_precision': _ComputePrecision.INTEGER,
            },
        },
    ]
    self._recipe_manager.load_quantization_recipe(full_quantization_config)

    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.BATCH_MATMUL, 'model/Dense10/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    # WEIGHT_ONLY check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.FLOAT)
    self.assertEqual(weight_tensor_config.num_bits, 8)

    # Dense will be overwritten by the last setting
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    # DRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)
    self.assertEqual(weight_tensor_config.num_bits, 4)

  def test_get_unsupported_op_fall_back_to_default(self):
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense/.*',
        operation_name=_TFLOpName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=17),
        ),
    )
    alg_key, _ = self._recipe_manager.get_quantization_configs(
        _TFLOpName.BATCH_MATMUL, 'model/Dense10/op'
    )
    # int17 is not supported, fall back to float.
    self.assertEqual(alg_key, _AlgorithmName.NO_QUANTIZE)

  def test_load_from_full_quantization_config_full_integer(self):
    full_quantization_config = [
        {
            'regex': '.*',
            'operation': '*',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'activation_tensor_config': {
                    'num_bits': 8,
                    'symmetric': False,
                    'granularity': _QuantGranularity.TENSORWISE,
                    'dtype': 'INT',
                },
                'weight_tensor_config': {
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': _QuantGranularity.TENSORWISE,
                    'dtype': 'INT',
                },
                # SRQ.
                'compute_precision': _ComputePrecision.INTEGER,
            },
        },
        {
            'regex': '.*',
            'operation': 'BATCH_MATMUL',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': _QuantGranularity.CHANNELWISE,
                },
                # WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
            },
        },
        {
            'regex': '.*/Dense/.*',
            'operation': '*',
            'algorithm_key': _AlgorithmName.MIN_MAX_UNIFORM_QUANT,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': 'INT',
                    'num_bits': 4,
                    'symmetric': False,
                    'granularity': _QuantGranularity.CHANNELWISE,
                },
                # DRQ.
                'compute_precision': _ComputePrecision.INTEGER,
            },
        },
    ]
    self._recipe_manager.load_quantization_recipe(full_quantization_config)

    # BMMs will be overridden to weight-only
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.BATCH_MATMUL, 'model/Dense10/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertIsNone(op_config.activation_tensor_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    # WEIGHT_ONLY check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.FLOAT)
    self.assertEqual(weight_tensor_config.num_bits, 8)

    # Dense will be overwritten by the last setting
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.FULLY_CONNECTED, 'model/Dense/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    self.assertIsNone(op_config.activation_tensor_config)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    # DRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)
    self.assertEqual(weight_tensor_config.num_bits, 4)

    # Other ops will have default quantization settings
    alg_key, op_config = self._recipe_manager.get_quantization_configs(
        _TFLOpName.CONV_2D, 'model/Dense11/op'
    )
    self.assertEqual(alg_key, _AlgorithmName.MIN_MAX_UNIFORM_QUANT)
    op_act_config = op_config.activation_tensor_config
    self.assertIsNotNone(op_act_config)
    self.assertEqual(op_act_config.num_bits, 8)
    weight_tensor_config = op_config.weight_tensor_config
    self.assertIsNotNone(weight_tensor_config)
    # SRQ check.
    self.assertEqual(op_config.compute_precision, _ComputePrecision.INTEGER)
    self.assertEqual(weight_tensor_config.num_bits, 8)

  def test_need_calibration_false(self):
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_1/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        ),
    )
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_2/.*',
        operation_name=_TFLOpName.CONV_2D,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )
    self.assertFalse(self._recipe_manager.need_calibration())

  def test_need_calibration_true(self):
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_1/.*',
        operation_name=_TFLOpName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        ),
    )
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_2/.*',
        operation_name=_TFLOpName.CONV_2D,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
        ),
    )
    self._recipe_manager.add_quantization_config(
        regex='.*/Dense_3/.*',
        operation_name=_TFLOpName.BATCH_MATMUL,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8),
            activation_tensor_config=_TensorQuantConfig(num_bits=8),
            compute_precision=_ComputePrecision.INTEGER,  # SRQ.
        ),
    )
    self.assertTrue(self._recipe_manager.need_calibration())


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import pathlib
from absl import flags
import numpy as np
import os
import absl.testing.absltest as absltest
from ai_edge_quantizer import model_validator
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_quantizer.utils import validation_utils

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('.')


class ComparisonResultTest(absltest.TestCase):

  def setUp(self):
    # TODO: b/358437395 - Remove this line once the bug is fixed.
    flags.FLAGS.mark_as_parsed()
    super().setUp()
    self.test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / 'tests/models/two_signatures.tflite'
    )
    self.test_model = tfl_flatbuffer_utils.get_model_buffer(
        self.test_model_path
    )
    self.test_quantized_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/two_signatures_a8w8.tflite',
    )
    self.test_quantized_model = tfl_flatbuffer_utils.get_model_buffer(
        self.test_quantized_model_path
    )
    self.test_data = {
        'add': {'add_x:0': 1e-3, 'Add/y': 0.25, 'PartitionedCall:0': 1e-3},
        'multiply': {
            'multiply_x:0': 1e-3,
            'Mul/y': 0.32,
            'PartitionedCall_1:0': 1e-2,
        },
    }
    self.test_dir = self.create_tempdir()
    self.comparison_result = model_validator.ComparisonResult(
        self.test_model, self.test_quantized_model
    )

  def test_add_new_signature_results_succeeds(self):
    for signature_key, test_result in self.test_data.items():
      self.comparison_result.add_new_signature_results(
          'mean_squared_difference',
          test_result,
          signature_key,
      )
    self.assertLen(
        self.comparison_result.available_signature_keys(), len(self.test_data)
    )

    for signature_key in self.test_data:
      signature_result = self.comparison_result.get_signature_comparison_result(
          signature_key
      )
      input_tensors = signature_result.input_tensors
      output_tensors = signature_result.output_tensors
      constant_tensors = signature_result.constant_tensors
      intermediate_tensors = signature_result.intermediate_tensors

      self.assertLen(input_tensors, 1)
      self.assertLen(output_tensors, 1)
      self.assertLen(constant_tensors, 1)
      self.assertEmpty(intermediate_tensors)

  def test_add_new_signature_results_fails_same_signature_key(self):
    self.comparison_result.add_new_signature_results(
        'mean_squared_difference',
        self.test_data['add'],
        'add',
    )
    error_message = 'add is already in the comparison_results.'
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self.comparison_result.add_new_signature_results(
          'mean_squared_difference',
          self.test_data['add'],
          'add',
      )

  def test_get_signature_comparison_result_fails_with_invalid_signature_key(
      self,
  ):
    self.comparison_result.add_new_signature_results(
        'mean_squared_difference',
        self.test_data['add'],
        'add',
    )
    error_message = 'multiply is not in the comparison_results.'
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self.comparison_result.get_signature_comparison_result('multiply')

  def test_get_all_tensor_results_succeeds(self):
    for signature_key, test_result in self.test_data.items():
      self.comparison_result.add_new_signature_results(
          'mean_squared_difference',
          test_result,
          signature_key,
      )
    all_tensor_results = self.comparison_result.get_all_tensor_results()
    self.assertLen(all_tensor_results, 6)
    self.assertIn('add_x:0', all_tensor_results)
    self.assertIn('Add/y', all_tensor_results)
    self.assertIn('PartitionedCall:0', all_tensor_results)
    self.assertIn('multiply_x:0', all_tensor_results)
    self.assertIn('Mul/y', all_tensor_results)
    self.assertIn('PartitionedCall_1:0', all_tensor_results)

  def test_save_comparison_result_succeeds(self):
    for signature_key, test_result in self.test_data.items():
      self.comparison_result.add_new_signature_results(
          'mean_squared_difference',
          test_result,
          signature_key,
      )
    model_name = 'test_model'
    self.comparison_result.save(self.test_dir.full_path, model_name)
    test_json_path = str(
        pathlib.Path(self.test_dir.full_path)
        / (model_name + '_comparison_result.json')
    )
    with open(test_json_path) as json_file:
      json_dict = json.load(json_file)

    # Check model size stats.
    self.assertIn('reduced_size_bytes', json_dict)
    self.assertEqual(
        json_dict['reduced_size_bytes'],
        len(self.test_model) - len(self.test_quantized_model),
    )
    self.assertIn('reduced_size_percentage', json_dict)
    self.assertEqual(
        json_dict['reduced_size_percentage'],
        (len(self.test_model) - len(self.test_quantized_model))
        / len(self.test_model)
        * 100,
    )

    for signature_key in self.test_data:
      self.assertIn(signature_key, json_dict)
      signature_result = json_dict[signature_key]
      self.assertIn('error_metric', signature_result)
      self.assertEqual(
          signature_result['error_metric'], 'mean_squared_difference'
      )
      self.assertIn('constant_tensors', signature_result)
      if signature_key == 'add':
        self.assertIn('Add/y', signature_result['constant_tensors'])
        self.assertNotIn('Mul/y', signature_result['constant_tensors'])
      elif signature_key == 'multiply':
        self.assertIn('Mul/y', signature_result['constant_tensors'])
        self.assertNotIn('Add/y', signature_result['constant_tensors'])

  def test_add_new_signature_results_validate_output_tensors_only(self):
    for signature_key, test_result in self.test_data.items():
      self.comparison_result.add_new_signature_results(
          'mean_squared_difference',
          test_result,
          signature_key,
          validate_output_tensors_only=True,
      )
    self.assertLen(
        self.comparison_result.available_signature_keys(), len(self.test_data)
    )

    for signature_key in self.test_data:
      signature_result = self.comparison_result.get_signature_comparison_result(
          signature_key
      )
      input_tensors = signature_result.input_tensors
      output_tensors = signature_result.output_tensors
      constant_tensors = signature_result.constant_tensors
      intermediate_tensors = signature_result.intermediate_tensors

      self.assertEmpty(input_tensors)
      self.assertLen(output_tensors, 1)
      self.assertEmpty(constant_tensors)
      self.assertEmpty(intermediate_tensors)


class ModelValidatorCompareTest(absltest.TestCase):

  def setUp(self):
    # TODO: b/358437395 - Remove this line once the bug is fixed.
    flags.FLAGS.mark_as_parsed()
    super().setUp()
    self.reference_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / 'tests/models/single_fc_bias.tflite'
    )
    self.target_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/single_fc_bias_sub_channel_weight_only_sym_weight.tflite',
    )
    self.reference_model = tfl_flatbuffer_utils.get_model_buffer(
        self.reference_model_path
    )
    self.target_model = tfl_flatbuffer_utils.get_model_buffer(
        self.target_model_path
    )
    self.signature_key = 'serving_default'  # single signature.
    self.test_data = tfl_interpreter_utils.create_random_normal_input_data(
        self.reference_model_path
    )
    self.test_dir = self.create_tempdir()

  def test_model_validator_compare(self):
    error_metric = 'mean_squared_difference'
    comparison_result = model_validator.compare_model(
        self.reference_model,
        self.target_model,
        self.test_data,
        error_metric,
        validation_utils.mean_squared_difference,
    )
    result = comparison_result.get_signature_comparison_result(
        self.signature_key
    )
    self.assertEqual(result.error_metric, 'mean_squared_difference')
    input_tensors = result.input_tensors
    output_tensors = result.output_tensors
    constant_tensors = result.constant_tensors
    intermediate_tensors = result.intermediate_tensors

    self.assertLen(input_tensors, 1)
    self.assertLen(output_tensors, 1)
    self.assertLen(constant_tensors, 2)
    self.assertEmpty(intermediate_tensors)

    self.assertAlmostEqual(input_tensors['serving_default_input_2:0'], 0)
    self.assertAlmostEqual(constant_tensors['arith.constant1'], 0)
    self.assertLess(output_tensors['StatefulPartitionedCall:0'], 1e-5)

  def test_model_validator_compare_validate_output_tensors_only(self):
    error_metric = 'mean_squared_difference'
    comparison_result = model_validator.compare_model(
        self.reference_model,
        self.target_model,
        self.test_data,
        error_metric,
        validation_utils.mean_squared_difference,
        validate_output_tensors_only=True,
    )
    result = comparison_result.get_signature_comparison_result(
        self.signature_key
    )
    self.assertEqual(result.error_metric, 'mean_squared_difference')
    input_tensors = result.input_tensors
    output_tensors = result.output_tensors
    constant_tensors = result.constant_tensors
    intermediate_tensors = result.intermediate_tensors

    self.assertEmpty(input_tensors)
    self.assertLen(output_tensors, 1)
    self.assertEmpty(constant_tensors)
    self.assertEmpty(intermediate_tensors)

    self.assertLess(output_tensors['StatefulPartitionedCall:0'], 1e-5)

  def test_create_json_for_model_explorer(self):
    error_metric = 'mean_squared_difference'
    comparison_result = model_validator.compare_model(
        self.reference_model,
        self.target_model,
        self.test_data,
        error_metric,
        validation_utils.mean_squared_difference,
    )
    mv_json = model_validator.create_json_for_model_explorer(
        comparison_result, [0, 1, 2, 3]
    )
    self.assertContainsSubset(
        '"thresholds": [{"value": 0, "bgColor": "rgb(200, 0, 0)"}, {"value":'
        ' 1, "bgColor": "rgb(200, 63, 0)"}, {"value": 2, "bgColor": "rgb(200,'
        ' 126, 0)"}, {"value": 3, "bgColor": "rgb(200, 189, 0)"}]',
        mv_json,
    )

  def test_create_json_for_model_explorer_no_thresholds(self):
    error_metric = 'mean_squared_difference'
    comparison_result = model_validator.compare_model(
        self.reference_model,
        self.target_model,
        self.test_data,
        error_metric,
        validation_utils.mean_squared_difference,
    )
    mv_json = model_validator.create_json_for_model_explorer(
        comparison_result, []
    )
    self.assertContainsSubset('"thresholds": []', mv_json)


class ModelValidatorMultiSignatureModelTest(absltest.TestCase):

  def setUp(self):
    # TODO: b/358437395 - Remove this line once the bug is fixed.
    flags.FLAGS.mark_as_parsed()
    super().setUp()
    self.reference_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / 'tests/models/two_signatures.tflite'
    )
    self.target_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/two_signatures_a8w8.tflite',
    )
    self.reference_model = tfl_flatbuffer_utils.get_model_buffer(
        self.reference_model_path
    )
    self.target_model = tfl_flatbuffer_utils.get_model_buffer(
        self.target_model_path
    )
    self.test_data = {
        'add': [{'x': np.array([2.0]).astype(np.float32)}],
        'multiply': [{'x': np.array([1.0]).astype(np.float32)}],
    }
    self.test_dir = self.create_tempdir()

  def test_model_validator_compare_succeeds(self):
    error_metric = 'mean_squared_difference'
    result = model_validator.compare_model(
        self.reference_model,
        self.target_model,
        self.test_data,
        error_metric,
        validation_utils.mean_squared_difference,
    )
    for signature_key in self.test_data:
      signature_result = result.get_signature_comparison_result(signature_key)
      input_tensors = signature_result.input_tensors
      output_tensors = signature_result.output_tensors
      constant_tensors = signature_result.constant_tensors
      intermediate_tensors = signature_result.intermediate_tensors

      self.assertLen(input_tensors, 1)
      self.assertLen(output_tensors, 1)
      self.assertLen(constant_tensors, 1)
      self.assertEmpty(intermediate_tensors)

      if signature_key == 'add':
        self.assertLess(input_tensors['add_x:0'], 1e-3)
        self.assertAlmostEqual(constant_tensors['Add/y'], 0)
        self.assertLess(output_tensors['PartitionedCall:0'], 1e-3)
      elif signature_key == 'multiply':
        self.assertLess(input_tensors['multiply_x:0'], 1e-3)
        self.assertAlmostEqual(constant_tensors['Mul/y'], 0)
        self.assertLess(output_tensors['PartitionedCall_1:0'], 1e-2)

  def test_create_json_for_model_explorer(self):
    error_metric = 'mean_squared_difference'
    comparison_result = model_validator.compare_model(
        self.reference_model,
        self.target_model,
        self.test_data,
        error_metric,
        validation_utils.mean_squared_difference,
    )
    thresholds = [0, 1, 2, 3]
    mv_json = model_validator.create_json_for_model_explorer(
        comparison_result, thresholds
    )
    self.assertContainsSubset(
        '"thresholds": [{"value": 0, "bgColor": "rgb(200, 0, 0)"}, {"value":'
        ' 1, "bgColor": "rgb(200, 63, 0)"}, {"value": 2, "bgColor": "rgb(200,'
        ' 126, 0)"}, {"value": 3, "bgColor": "rgb(200, 189, 0)"}]',
        mv_json,
    )

  def test_create_json_for_model_explorer_no_thresholds(self):
    error_metric = 'mean_squared_difference'
    comparison_result = model_validator.compare_model(
        self.reference_model,
        self.target_model,
        self.test_data,
        error_metric,
        validation_utils.mean_squared_difference,
    )
    mv_json = model_validator.create_json_for_model_explorer(
        comparison_result, []
    )
    self.assertContainsSubset('"thresholds": []', mv_json)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility functions for graph transformations."""

import copy
import dataclasses
from typing import Optional, Union

import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import


@dataclasses.dataclass
class TransformationInput:
  """Standard input for a graph transformation.

  Attributes:
    tensor_id: the tensor index to insert dequant op after
    op_codes: list of operatorCode in the model, if dequantize op doesn't exist,
      we need to insert the op code into the list
    buffers: list of buffer in the original TFlite model for buffer quantization
    subgraph: flatbuffer subgraph object which the tensor resides.
    producer: op id for the producer of the tensor.
    consumers: op ids for consumers of the new dequant op.
    quant_params: quantization parameters to be applied on the orignal tensor
  """

  tensor_id: int
  op_codes: list[schema_py_generated.OperatorCodeT]
  buffers: list[schema_py_generated.BufferT]
  subgraph: schema_py_generated.SubGraphT
  producer: int
  consumers: list[int]
  quant_params: Union[qtyping.UniformQuantParams, qtyping.NonLinearQuantParams]


def add_op_code(
    op_code: schema_py_generated.OperatorCodeT,
    model_op_codes: list[schema_py_generated.OperatorCodeT],
    custom_op_name: Optional[str] = None,
) -> int:
  """Add an op code into a model if it's not present.

  Args:
    op_code: The op code to be added.
    model_op_codes: The op codes of the model.
    custom_op_name: The custom string of the op code. If None, the op code will
      be added as a builtin op code.

  Returns:
    The index of the op code in the model.
  """
  if (
      op_code == schema_py_generated.BuiltinOperator.CUSTOM
      and custom_op_name is None
  ):
    raise ValueError('Custom string is required for custom op code.')

  for i, model_op_code in enumerate(model_op_codes):
    # If the model already has the op code, just return the index.
    if model_op_code.builtinCode == op_code:
      if custom_op_name is not None:
        if model_op_code.customCode == custom_op_name:
          return i
      else:
        # Built-in op
        return i

  model_op_codes.append(schema_py_generated.OperatorCodeT())
  model_op_codes[-1].builtinCode = op_code
  if custom_op_name is not None:
    model_op_codes[-1].customCode = custom_op_name
  return len(model_op_codes) - 1


def get_constant_buffer(
    data: np.ndarray,
    buffers: list[schema_py_generated.BufferT],
    force_duplicate_buffer: bool = False,
) -> int:
  """Get the index of the constant buffer that contains the given data.

  creating new buffer if provided data is not found in buffers list.

  Args:
    data: The data of the new tensor.
    buffers: The buffers of the model.
    force_duplicate_buffer: Whether to add a new buffer even if the same buffer
      already exists.

  Returns:
    The index of the new buffer in the model.
  """

  if isinstance(data, np.ndarray):
    # in the case where the data is passed from quantization_params.
    new_data = np.frombuffer(data.tobytes(), dtype=np.uint8).flatten()
  elif isinstance(data, bytes):
    # in the case where the data is coming from duplicating buffers, we need to
    # make a copy of the data to avoid having two buffers pointing to the same
    # data.
    new_data = copy.deepcopy(data)
  else:
    raise ValueError('data passed in must be either np.ndarray or bytes.')
  # TODO: b/417811116 - we should make this more efficient.
  if not force_duplicate_buffer:
    for index, buffer in enumerate(buffers):
      if np.array_equal(buffer.data, new_data):
        return index
  new_buffer = schema_py_generated.BufferT()
  new_buffer.data = new_data
  new_buffer.offset = 0
  new_buffer.size = 0
  new_buffer_id = len(buffers)
  buffers.append(new_buffer)

  return new_buffer_id


def add_new_constant_tensor(
    tensor_name: str,
    data: np.ndarray,
    tensor_type: schema_py_generated.TensorType,
    subgraph: schema_py_generated.SubGraphT,
    buffers: list[schema_py_generated.BufferT],
    tensor_shape: Optional[list[int]] = None,
    force_duplicate_buffer: bool = False,
    quantization: schema_py_generated.QuantizationParametersT | None = None,
) -> int:
  """Add a new constant tensor to the model.

  Args:
    tensor_name: The name of the new tensor.
    data: The data of the new tensor.
    tensor_type: The type of the new tensor.
    subgraph: The subgraph where the new tensor is added.
    buffers: The buffers of the model.
    tensor_shape: The shape of the new tensor. If not provided, the shape of the
      data will be used.
    force_duplicate_buffer: Whether to add a new buffer even if the same buffer
      already exists.
    quantization: Optional `QuantizationParametersT` describing the quantization
      of this tensor.

  Returns:
    The index of the new tensor in the subgraph.
  """
  new_buffer_id = get_constant_buffer(data, buffers, force_duplicate_buffer)

  new_tensor = schema_py_generated.TensorT()
  if tensor_shape is None:
    tensor_shape = data.shape
  new_tensor.shape = tensor_shape
  new_tensor.buffer = new_buffer_id
  new_tensor.type = tensor_type
  new_tensor.name = tensor_name
  new_tensor.quantization = quantization
  new_tensor_id = len(subgraph.tensors)
  subgraph.tensors.append(new_tensor)
  return new_tensor_id


def add_new_activation_tensor(
    tensor_name: str,
    shape: list[int],
    tensor_type: schema_py_generated.TensorType,
    subgraph: schema_py_generated.SubGraphT,
    quantization: schema_py_generated.QuantizationParametersT | None = None,
) -> int:
  """Add a new activation tensor to the model.

  Args:
    tensor_name: The name of the new tensor.
    shape: The shape of the new tensor.
    tensor_type: The type of the new tensor.
    subgraph: The subgraph where the new tensor is added.
    quantization: Optional `QuantizationParametersT` describing the quantization
      of this tensor.

  Returns:
    The index of the new tensor in the subgraph.
  """
  new_tensor = schema_py_generated.TensorT()
  # If there's a dynamic shape, we need to read from the shapeSignature field
  # instead of shape. Shape should contain just 1 for the dynamic dimension but
  # shapeSignature should contain the true shape.
  if -1 in shape:
    new_tensor.shapeSignature = shape
    new_tensor.shape = [1 if i == -1 else i for i in shape]
  else:
    new_tensor.shape = shape
  new_tensor.type = tensor_type
  new_tensor.name = tensor_name
  new_tensor.quantization = quantization
  new_tensor.buffer = 0
  new_tensor_id = len(subgraph.tensors)
  subgraph.tensors.append(new_tensor)
  return new_tensor_id


def raise_deprecated_error(_: TransformationInput):
  raise NotImplementedError(
      'This transformation is deprecated. Please contact AI Edge Quantizer team'
      ' if you see this error.'
  )


def pack_data(bitwidth: int, flattened_data: np.ndarray) -> np.ndarray:
  """Pack the data to the corresponding bit width.

  Currently only support 4 bits. If no packing is needed, the original data is
  returned.

  Args:
    bitwidth: Bit width from NonLinearQuantParams.
    flattened_data: The data to be packed.

  Returns:
    Packed data.
  """
  if bitwidth == 4:
    flattened_data = np.bitwise_and(flattened_data.astype(np.uint8), 0x0F)
    even_data = flattened_data[::2]
    odd_data = np.left_shift(flattened_data[1::2], 4)
    if odd_data.shape[0] == even_data.shape[0] - 1:
      odd_data = np.pad(odd_data, (0, 1), constant_values=0)
    return np.bitwise_or(even_data, odd_data)
  else:
    return flattened_data


def get_producer_schema_op_id(
    transformation: TransformationInput,
) -> int:
  """Checks if the tensor's producer matches the given op.

  Args:
    transformation: The transformation input to check the producer of.

  Returns:
    The schema op id of the producer op. E.g.
    schema_py_generated.BuiltinOperator.FULLY_CONNECTED.
  """
  if transformation.producer == -1:
    return False
  else:
    return (
        transformation.op_codes[
            transformation.subgraph.operators[
                transformation.producer
            ].opcodeIndex
        ].builtinCode
    )


def get_schema_op_id(
    transformation: TransformationInput, op_id: int
) -> bool:
  """Returns the schema op id of the given op.

  Args:
    transformation: The transformation input to check the consumers of.
    op_id: The op id in the list of operators to check for.

  Returns:
    The schema op id of the given op.
  """
  return (
      transformation.op_codes[
          transformation.subgraph.operators[op_id].opcodeIndex
      ].builtinCode
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import octav
from ai_edge_quantizer.algorithms.uniform_quantize.op_architecture_tests import test_utils as op_test_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile(
    "../../../tests/models"
)
_TFLOpName = qtyping.TFLOperationName
_ComputePrecision = qtyping.ComputePrecision
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_QuantTransformation = qtyping.QuantTransformation
_OpTestInfo = op_test_utils.OpTestInfo
_QuantGranularity = qtyping.QuantGranularity


class EmbeddingLookupTest(op_test_utils.BaseQuantizeTest):

  def setUp(self):
    super().setUp()
    np.random.seed(666)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "embedding_lookup.tflite"
    )
    self._op_test_info = _OpTestInfo(
        test_model=tfl_flatbuffer_utils.read_model(self._test_model_path),
        op_tensor_names={},
        input_range=(np.array([[-10]]), np.array([[8]])),
        output_range=(np.array([[10]]), np.array([[88]])),
    )
    # The test model has one subgraph for now.
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._op_test_info.test_model.subgraphs[0].tensors,
        buffers=self._op_test_info.test_model.buffers,
    )

  @parameterized.product(
      num_bits_weight=(4, 8),
      granularity=(
          qtyping.QuantGranularity.CHANNELWISE,
          qtyping.QuantGranularity.TENSORWISE,
      ),
      # get_tensor_quant_params_func, symmetric_weight
      algos=(
          (naive_min_max_quantize.get_tensor_quant_params, True),
          (naive_min_max_quantize.get_tensor_quant_params, False),
          (octav.get_tensor_quant_params, True),
      ),
      test_case=(
          # Tuple holds compute precision and whether to use srq and explicit
          # dequantize.
          (_ComputePrecision.FLOAT, True),
          (_ComputePrecision.INTEGER, False),
      ),
  )
  def test_embedding_lookup_succeeds(
      self,
      num_bits_weight,
      granularity,
      algos,
      test_case,
  ):
    get_quant_params_func, symmetric_weight = algos
    compute_precision, explicit_dequantize = test_case

    # Read from Model Explorer.
    subgraph0 = self._op_test_info.test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]
    op_tensor_names = {}
    op_tensor_names["weight"] = (
        "jit(export_func)/jit(main)/...y,yz->...z/dot_general;jit(export_func)/jit(main)/jit(_one_hot)/eq;jit(export_func)/jit(main)/jit(_one_hot)/convert_element_type"
    )
    op_tensor_names["input"] = "lookup"
    op_tensor_names["output"] = "Identity_1"
    self._op_test_info.op_tensor_names = op_tensor_names
    self._op_test_info.quantized_dimension = 0

    op_info = qtyping.OpInfo(
        op=op,
        op_name=qtyping.TFLOperationName.EMBEDDING_LOOKUP,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=num_bits_weight,
                symmetric=symmetric_weight,
                granularity=granularity,
            ),
            compute_precision=compute_precision,
            explicit_dequantize=explicit_dequantize,
        ),
    )
    # TODO: b/335913710 - Rename the test function.
    self._test_fc_bmm_conv(
        op_info,
        self._graph_info,
        self._op_test_info,
        common_quantize.materialize_embedding_lookup,
        get_quant_params_func,
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements the MSE quantization."""

import dataclasses
from typing import Any, Optional
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_quantizer.algorithms.utils import common_utils

ALGORITHM_KEY = "MSE"

# Coefficients from offline numeric analysis.
_MSE_QUANT_MULS = {
    8: 0.05408,
    4: 0.37755,
}


def get_tensor_quant_params(
    op_info: qtyping.OpInfo,
    tensor_quant_config: qtyping.TensorQuantizationConfig,
    tensor_content: Optional[np.ndarray] = None,
    tensor_qsv: Optional[dict[str, Any]] = None,
) -> qtyping.UniformQuantParams:
  """Returns the quantization parameters for a tensor.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    tensor_quant_config: The quantization config for the tensor.
    tensor_content: The content of the tensor. When None, it means the tensor is
      not a weight tensor (e.g. static quantization) so we fallback to using
      naive_min_max_quantize.
    tensor_qsv: A dictionary containing the min/max of the tensor.

  Raises:
    ValueError: If the blockwise quantization is requested.
    ValueError: If the asymmetric quantization is requested.
    ValueError: `tensor_qsv` must contain min/max values, or `tensor_content`
      must be provided so that they can be inferred.
  """
  if uniform_quantize_tensor.is_blockwise(tensor_quant_config.granularity):
    raise ValueError(
        "Blockwise quantization is not supported for MSE quantization."
    )

  # Fallback to naive_min_max_quantize.py for non-weight tensors.
  if tensor_content is None:
    return naive_min_max_quantize.get_tensor_quant_params(
        op_info, tensor_quant_config, tensor_content, tensor_qsv
    )

  if not tensor_quant_config.symmetric:
    raise ValueError(
        f"Unsupported symmetry: {tensor_quant_config.symmetric}. MSE"
        " supports symmetric quantization only for now."
    )

  if not tensor_qsv:
    # We need min/max to calculate quantization parameters, which
    # should be collected during the calibration process. However,
    # weight-only and DRQ do not require calibration, thus it is
    # possible that this information is missing here. In that case we
    # collect min/max on the spot.
    tensor_min_max = common_quantize.init_tensor_min_max(
        tensor_content,
        op_info,
    )
  else:
    tensor_min_max = tensor_qsv

  if "min" not in tensor_min_max or "max" not in tensor_min_max:
    raise ValueError(
        "min and max must be provided to produce tensor quantization"
        " parameters. Check if the correct calibration results are passed into"
        " the ParamsGenerator."
    )

  quantized_dim = common_utils.get_weight_quantized_dim(
      op_info, tensor_content, tensor_quant_config.granularity
  )

  reshaped_data = tensor_content
  reduce_dims = common_utils.get_reduce_dims(
      quantized_dim, tensor_content.shape
  )

  multiplier = _MSE_QUANT_MULS[tensor_quant_config.num_bits]
  scale = multiplier * np.sqrt(
      np.mean(reshaped_data**2, axis=reduce_dims, keepdims=True)
  )
  zp = np.zeros_like(scale, dtype=np.int32)

  quant_params = qtyping.UniformQuantParams(
      scale=scale,
      zero_point=zp,
      num_bits=tensor_quant_config.num_bits,
      symmetric=tensor_quant_config.symmetric,
      quantized_dimension=quantized_dim,
      block_size=uniform_quantize_tensor.extract_block_size_from_granularity(
          tensor_quant_config.granularity
      ),
  )

  quantized_vars = uniform_quantize_tensor.uniform_quantize(
      tensor_content,
      quant_params,
      uniform_quantize_tensor.is_blockwise(tensor_quant_config.granularity),
  )

  return dataclasses.replace(quant_params, quantized_data=quantized_vars)



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import default_policy
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("../../tests/models")
_TFLOpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig


class CommonQuantizeTest(parameterized.TestCase):
  """Tests for general quantize functions."""

  def setUp(self):
    super().setUp()
    np.random.seed(666)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "conv_fc_mnist.tflite"
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    # The test model has one subgraph for now.
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._test_model.subgraphs[0].tensors,
        buffers=self._test_model.buffers,
    )
    self._tensor_name_to_qsv = {}

  def test_check_op_quantization_config_with_negative_min_weight_elements_raises_error(
      self,
  ):
    op_quant_config = qtyping.OpQuantizationConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=qtyping.ComputePrecision.INTEGER,  # DRQ.
        min_weight_elements=-1,
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError,
        lambda err: "min_weight_elements must be non-negative" in str(err),
    ):
      common_quantize.check_op_quantization_config(
          _TFLOpName.FULLY_CONNECTED,
          op_quant_config,
          default_policy.DEFAULT_CONFIG_CHECK_POLICY,
      )

  def test_reshape_data_for_blockwise_raises_error_when_quantized_dim_not_divisible_by_block_size(
      self,
  ):
    tensor_data = np.ones((24, 128), dtype=np.float32)
    block_size = 256
    quantized_dim = 1
    with self.assertRaisesWithPredicateMatch(
        ValueError,
        lambda err: (
            "Tensor quantization dimension must be divisible by block"
            " size for blockwise quantization."
        )
        in str(err),
    ):
      common_quantize._reshape_data_for_blockwise(
          tensor_data, quantized_dim, block_size
      )

  def test_reshape_data_for_blockwise_returns_correct_values(self):
    tensor_data = np.ones((24, 128), dtype=np.float32)
    block_size = 32
    quantized_dim = 1
    new_tensor_data, reduce_dim = common_quantize._reshape_data_for_blockwise(
        tensor_data, quantized_dim, block_size
    )
    self.assertEqual(new_tensor_data.shape, (24, 4, 32))
    self.assertEqual(reduce_dim, 2)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for instruction_generator."""

from collections.abc import Sequence
import pathlib
from typing import Optional

import numpy as np

from absl.testing import parameterized
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import transformation_instruction_generator as instruction_generator
from ai_edge_quantizer.utils import test_utils

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile(".")

_QTransf = qtyping.QuantTransformation


class InstructionGeneratorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name="second_index_test",
          param1=qtyping.OpToTensorParams(
              0,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
                  qtyping.QuantTransformation.ADD_DEQUANTIZE,
              ],
              qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
          ),
          param2=qtyping.OpToTensorParams(
              2,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
                  qtyping.QuantTransformation.ADD_DEQUANTIZE,
              ],
              qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
          ),
          index=1,
          expected=True,
      ),
      dict(
          testcase_name="different_trans_length_test",
          param1=qtyping.OpToTensorParams(
              0,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
              ],
              qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
          ),
          param2=qtyping.OpToTensorParams(
              2,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
                  qtyping.QuantTransformation.ADD_DEQUANTIZE,
              ],
              qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
          ),
          index=1,
          expected=False,
      ),
      dict(
          testcase_name="different_trans_length_test2",
          param1=qtyping.OpToTensorParams(
              0,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
                  qtyping.QuantTransformation.ADD_DEQUANTIZE,
              ],
              qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
          ),
          param2=qtyping.OpToTensorParams(
              2,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
              ],
              qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
          ),
          index=1,
          expected=False,
      ),
      dict(
          testcase_name="test_unmatched_transforamtions",
          param1=qtyping.OpToTensorParams(
              0,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
              ],
              qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0])),
          ),
          param2=qtyping.OpToTensorParams(
              2,
              [
                  qtyping.QuantTransformation.ADD_QUANTIZE,
                  qtyping.QuantTransformation.ADD_DEQUANTIZE,
              ],
              qtyping.UniformQuantParams(
                  16, None, np.array([1]), np.array([0])
              ),
          ),
          index=0,
          expected=False,
      ),
  )
  def test_check_horizontal_optimization(self, param1, param2, index, expected):
    got = instruction_generator.check_horizontal_optimization(
        param1=param1, param2=param2, index=index
    )
    self.assertEqual(expected, got)

  @parameterized.named_parameters(
      dict(
          testcase_name="test_success",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          expected=True,
      ),
      dict(
          testcase_name="test_wrong_transformation",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          expected=False,
      ),
      dict(
          testcase_name="test_wrong_parameters",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  16, None, np.array([1]), np.array([0])
              ),
          ),
          expected=False,
      ),
  )
  def test_check_dq_q_elimination(self, producer_inst, consumer_inst, expected):
    got = instruction_generator.check_dq_q_elimination(
        producer_inst=producer_inst, consumer_inst=consumer_inst
    )
    self.assertEqual(expected, got)

  @parameterized.named_parameters(
      dict(
          testcase_name="test_success",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  16, None, np.array([1]), np.array([0])
              ),
          ),
          expected=True,
      ),
      dict(
          testcase_name="test_wrong_transformation",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          expected=False,
      ),
      dict(
          testcase_name="test_wrong_parameters",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          expected=False,
      ),
  )
  def test_check_replace_dq_q_with_rq(
      self, producer_inst, consumer_inst, expected
  ):
    got = instruction_generator.check_replace_dq_q_with_rq(
        producer_inst=producer_inst, consumer_inst=consumer_inst
    )
    self.assertEqual(expected, got)

  @parameterized.named_parameters(
      dict(
          testcase_name="test_elimination_success",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.NO_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          expected=True,
      ),
      dict(
          testcase_name="test_wrong_transformation1",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          expected=False,
      ),
      dict(
          testcase_name="test_wrong_transformation2",
          producer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_inst=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.NO_QUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          expected=False,
      ),
  )
  def test_check_dq_no_quant_elimination(
      self, producer_inst, consumer_inst, expected
  ):
    got = instruction_generator.check_dq_no_quant_elimination(
        producer_inst, consumer_inst
    )
    self.assertEqual(expected, got)

  @parameterized.named_parameters(
      dict(
          testcase_name="test_empty_consumer",
          producer_trans_rule=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_trans_rule=[],
          expected=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              )
          ],
      ),
      dict(
          testcase_name="test_no_vertical_trans",
          producer_trans_rule=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[0, np.array([1]), 2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_trans_rule=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[0],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          expected=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[0, np.array([1]), 2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[0],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
      ),
      dict(
          testcase_name="test_vertical_trans_with_mix_output",
          producer_trans_rule=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[0, np.array([1]), 2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_trans_rule=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[0],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      16, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          expected=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.QUANTIZE_TENSOR,
                  tensor_id=1,
                  producer=0,
                  consumers=[0],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.QUANTIZE_TENSOR,
                  tensor_id=1,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      16, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
      ),
      dict(
          testcase_name="test_multi_match",
          producer_trans_rule=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[0, 1, 2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_trans_rule=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[0, 1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          expected=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.QUANTIZE_TENSOR,
                  tensor_id=1,
                  producer=0,
                  consumers=[0, 1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
      ),
      dict(
          testcase_name="test_dequant_no_quant_elimination_succeeds",
          producer_trans_rule=qtyping.TransformationInst(
              transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
              tensor_id=1,
              producer=0,
              consumers=[0, 1, 2],
              parameters=qtyping.UniformQuantParams(
                  8, None, np.array([1]), np.array([0])
              ),
          ),
          consumer_trans_rule=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.NO_QUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[0, 1, 2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          expected=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[0, 1, 2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
      ),
  )
  def test_apply_vertical_optimization(
      self, producer_trans_rule, consumer_trans_rule, expected
  ):
    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/single_fc_bias.tflite"
        )
    )
    got = ins_gen._apply_vertical_optimization(
        producer_trans_rule, consumer_trans_rule
    )
    self.assertEqual(expected, got)

  @parameterized.named_parameters(
      dict(testcase_name="test_empty_consumer", param={}, expected=[]),
      dict(
          testcase_name="test_multi_level_grouping",
          param=qtyping.TensorTransformationParams(
              "tfl.quantize",
              qtyping.OpToTensorParams(
                  subgraph_op_id=0,
                  transformations=[qtyping.QuantTransformation.ADD_DEQUANTIZE],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              [
                  qtyping.OpToTensorParams(
                      subgraph_op_id=1,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=2,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE,
                          qtyping.QuantTransformation.ADD_DEQUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=3,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE,
                          qtyping.QuantTransformation.ADD_DEQUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=4,
                      transformations=[
                          qtyping.QuantTransformation.NO_QUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
              ],
          ),
          expected=[
              [{0, 1, 2, 3}],
              [{0, 1, 2}, {3}],
              [{1, 2}],
          ],
      ),
  )
  def test_group_consumer_transformations(self, param, expected):
    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/single_fc_bias.tflite"
        )
    )
    got = ins_gen._group_consumer_transformations(param)
    self.assertEqual(expected, got)

  @parameterized.named_parameters(
      dict(
          testcase_name="test_empty_input",
          consumer_group=[],
          param=qtyping.TensorTransformationParams(
              "arg0",
              None,
              [],
          ),
          expected=[],
      ),
      dict(
          testcase_name="test_multi_level_grouping",
          consumer_group=[
              [{0, 1, 2, 3}],
              [{0, 1, 2}, {3}],
              [{1, 2}],
          ],
          param=qtyping.TensorTransformationParams(
              "tfl.quantize",
              qtyping.OpToTensorParams(
                  subgraph_op_id=0,
                  transformations=[qtyping.QuantTransformation.ADD_DEQUANTIZE],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              [
                  qtyping.OpToTensorParams(
                      subgraph_op_id=1,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=2,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE,
                          qtyping.QuantTransformation.ADD_DEQUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=3,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE,
                          qtyping.QuantTransformation.ADD_DEQUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=4,
                      transformations=[
                          qtyping.QuantTransformation.NO_QUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
              ],
          ),
          expected=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[1, 2, 3],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.NO_QUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[4],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
      ),
  )
  def test_produce_transformation_for_vertical_opt(
      self, consumer_group, param, expected
  ):
    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/insert_dequant_test.tflite"
        )
    )
    got = ins_gen._produce_transformation_for_vertical_opt(
        consumer_group, param
    )
    self.assertEqual(expected, got)

  @parameterized.named_parameters(
      dict(
          testcase_name="test_empty_input",
          consumer_group=[],
          param=qtyping.TensorTransformationParams(
              "arg0",
              None,
              [],
          ),
          expected=[],
      ),
      dict(
          testcase_name="test_multi_level_grouping",
          consumer_group=[
              [{0, 1, 2, 3}],
              [{0, 1, 2}, {3}],
              [{1, 2}],
          ],
          param=qtyping.TensorTransformationParams(
              "tfl.quantize",
              qtyping.OpToTensorParams(
                  subgraph_op_id=0,
                  transformations=[qtyping.QuantTransformation.ADD_DEQUANTIZE],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              [
                  qtyping.OpToTensorParams(
                      subgraph_op_id=1,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=2,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE,
                          qtyping.QuantTransformation.ADD_DEQUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=3,
                      transformations=[
                          qtyping.QuantTransformation.ADD_QUANTIZE,
                          qtyping.QuantTransformation.ADD_DEQUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
                  qtyping.OpToTensorParams(
                      subgraph_op_id=4,
                      transformations=[
                          qtyping.QuantTransformation.NO_QUANTIZE,
                      ],
                      parameters=qtyping.UniformQuantParams(
                          8, None, np.array([1]), np.array([0])
                      ),
                  ),
              ],
          ),
          expected=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=1,
                  producer=0,
                  consumers=[2, 3],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
      ),
  )
  def test_produce_customer_transformations_unavailable_for_vertical_opt(
      self, consumer_group, param, expected
  ):
    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/insert_dequant_test.tflite"
        )
    )
    got = (
        ins_gen._produce_consumer_transformations_unavailable_for_vertical_opt(
            consumer_group, param
        )
    )
    self.assertEqual(expected, got)

  def test_empty_param(self):
    """test the capability to handle empty params."""
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/single_fc_bias.tflite"
    )
    quant_parameters = {}
    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        test_model_path
    )
    instructions = ins_gen.quant_params_to_transformation_insts(
        quant_parameters
    )
    self.assertEmpty(instructions)

  def test_generate_instruction_for_single_fc_bias(self):
    """test the capability to run multiple tensor infos."""
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/single_fc_bias.tflite"
    )
    quant_parameters = {}
    quant_parameters["serving_default_input_2:0"] = (
        qtyping.TensorTransformationParams(
            "serving_default_input_2:0",
            None,
            [
                qtyping.OpToTensorParams(
                    0,
                    [qtyping.QuantTransformation.ADD_QUANTIZE],
                    qtyping.UniformQuantParams(
                        8, None, np.array([1]), np.array([0])
                    ),
                )
            ],
        )
    )

    quant_parameters["StatefulPartitionedCall:0"] = (
        qtyping.TensorTransformationParams(
            "StatefulPartitionedCall:0",
            qtyping.OpToTensorParams(
                0,
                [
                    qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    qtyping.QuantTransformation.ADD_QUANTIZE,
                ],
                qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
            [],
        )
    )

    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        test_model_path
    )
    instructions = ins_gen.quant_params_to_transformation_insts(
        quant_parameters
    )
    input_transformation = qtyping.TensorTransformationInsts(
        tensor_name="serving_default_input_2:0",
        subgraph_id=0,
        instructions=[
            qtyping.TransformationInst(
                transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                tensor_id=0,
                producer=-1,  # input tensor is the subgraph input
                consumers=[0],  # consumed by node 0
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            )
        ],
    )
    output_transformation = qtyping.TensorTransformationInsts(
        tensor_name="StatefulPartitionedCall:0",
        subgraph_id=0,
        instructions=[
            qtyping.TransformationInst(
                transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                tensor_id=3,
                producer=0,
                consumers=[-1],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
            qtyping.TransformationInst(
                transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                tensor_id=3,
                producer=0,
                consumers=[-1],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
        ],
    )
    self.assertLen(instructions, 2)
    self.assertEqual(
        instructions["serving_default_input_2:0"], input_transformation
    )
    self.assertEqual(
        instructions["StatefulPartitionedCall:0"], output_transformation
    )

  def test_raise_error_on_no_quant_conflict(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/insert_dequant_test.tflite"
    )
    quant_parameters = {}
    quant_parameters["tfl.quantize"] = qtyping.TensorTransformationParams(
        "tfl.quantize",
        None,
        [
            qtyping.OpToTensorParams(
                subgraph_op_id=1,
                transformations=[qtyping.QuantTransformation.QUANTIZE_TENSOR],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
            qtyping.OpToTensorParams(
                subgraph_op_id=2,
                transformations=[qtyping.QuantTransformation.NO_QUANTIZE],
                parameters=None,
            ),
        ],
    )
    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        test_model_path
    )
    with self.assertRaisesRegex(
        ValueError, "can not be both quantized and unquantized"
    ):
      ins_gen.quant_params_to_transformation_insts(quant_parameters)

  def test_generate_instruction_for_branching(self):
    """test horizontal and vertial optimization on a graph with multi branch."""
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/insert_dequant_test.tflite"
    )
    quant_parameters = {}
    quant_parameters["tfl.quantize"] = qtyping.TensorTransformationParams(
        "tfl.quantize",
        qtyping.OpToTensorParams(
            subgraph_op_id=0,
            transformations=[qtyping.QuantTransformation.ADD_DEQUANTIZE],
            parameters=qtyping.UniformQuantParams(
                8, None, np.array([1]), np.array([0])
            ),
        ),
        [
            qtyping.OpToTensorParams(
                subgraph_op_id=1,
                transformations=[qtyping.QuantTransformation.ADD_QUANTIZE],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
            qtyping.OpToTensorParams(
                subgraph_op_id=2,
                transformations=[
                    qtyping.QuantTransformation.ADD_QUANTIZE,
                    qtyping.QuantTransformation.ADD_DEQUANTIZE,
                ],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
        ],
    )
    ins_gen = instruction_generator.TransformationInstructionsGenerator(
        test_model_path
    )
    instructions = ins_gen.quant_params_to_transformation_insts(
        quant_parameters
    )
    expected_instructions = qtyping.TensorTransformationInsts(
        tensor_name="tfl.quantize",
        subgraph_id=0,
        instructions=[
            qtyping.TransformationInst(
                transformation=qtyping.QuantTransformation.QUANTIZE_TENSOR,
                tensor_id=1,
                producer=0,
                consumers=[1, 2],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
            qtyping.TransformationInst(
                transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                tensor_id=1,
                producer=0,
                consumers=[2],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
        ],
    )
    self.assertLen(instructions, 1)
    self.assertEqual(instructions["tfl.quantize"], expected_instructions)

  def test_instruction_generator_keeps_buffer_duplication_as_first_transformation(
      self,
  ):
    test_tensor_name = "test_tensor"

    dummy_quant_params = qtyping.UniformQuantParams(
        8, None, np.array([1]), np.array([0])
    )
    consumer_params_1 = qtyping.OpToTensorParams(
        subgraph_op_id=0,
        transformations=[
            qtyping.QuantTransformation.DUPLICATE_BUFFER,
            qtyping.QuantTransformation.ADD_QUANTIZE,
        ],
        parameters=dummy_quant_params,
    )
    consumer_params_2 = qtyping.OpToTensorParams(
        subgraph_op_id=2,
        transformations=[
            qtyping.QuantTransformation.DUPLICATE_BUFFER,
            qtyping.QuantTransformation.ADD_QUANTIZE,
            qtyping.QuantTransformation.ADD_DEQUANTIZE,
        ],
        parameters=dummy_quant_params,
    )

    quant_parameters = {
        test_tensor_name: qtyping.TensorTransformationParams(
            tensor_name=test_tensor_name,
            producer=None,
            consumers=[consumer_params_1, consumer_params_2],
        ),
    }
    instruction_gen = (
        instruction_generator.TransformationInstructionsGenerator()
    )
    # _tensor_name_to_graph_info has to have an entry for the test tensor for
    # `quant_params_to_transformation_insts` to work. But the values do not
    # matter for this test.
    instruction_gen._tensor_name_to_graph_info[test_tensor_name] = (
        instruction_generator.TransformationInstructionsGenerator.TensorGraphInfo(
            tensor_id=1,
            subgraph_id=0,
            producer=0,
            consumers=[2],
        )
    )
    instructions = instruction_gen.quant_params_to_transformation_insts(
        quant_parameters
    )
    self.assertLen(instructions, 1)
    instructions = instructions[test_tensor_name].instructions
    self.assertGreater(len(instructions), 1)
    self.assertEqual(instructions[0].transformation, _QTransf.DUPLICATE_BUFFER)
    self.assertNotIn(_QTransf.DUPLICATE_BUFFER, instructions[1:])

  def _get_test_instruction(self, transformation, consumers=None):
    if consumers is None:
      consumers = []
    return qtyping.TransformationInst(
        transformation=transformation,
        consumers=consumers,
        # Dummy values below.
        tensor_id=0,
        producer=None,
        parameters=None,
    )

  def test__remove_last_tensor_duplication_succeeds(self):
    tensor_instructions = qtyping.TensorTransformationInsts(
        tensor_name="test_tensor",
        subgraph_id=0,
        instructions=[
            self._get_test_instruction(_QTransf.DUPLICATE_TENSOR),
            self._get_test_instruction(_QTransf.ADD_QUANTIZE),
            self._get_test_instruction(_QTransf.DUPLICATE_TENSOR),
            self._get_test_instruction(_QTransf.ADD_DEQUANTIZE),
        ],
    )
    instruction_gen = (
        instruction_generator.TransformationInstructionsGenerator()
    )
    instruction_gen._remove_last_tensor_duplication(tensor_instructions)

    self.assertLen(tensor_instructions.instructions, 3)
    expected_transformations = [
        _QTransf.DUPLICATE_TENSOR,
        _QTransf.ADD_QUANTIZE,
        _QTransf.ADD_DEQUANTIZE,
    ]
    got_transformations = [
        instruction.transformation
        for instruction in tensor_instructions.instructions
    ]
    self.assertEqual(got_transformations, expected_transformations)

  def test__remove_unnecessary_buffer_duplication_succeeds(
      self,
  ):
    instructions = [
        self._get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[1]),
        self._get_test_instruction(_QTransf.DUPLICATE_BUFFER, consumers=[1]),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE),
        self._get_test_instruction(_QTransf.DUPLICATE_BUFFER, consumers=[1]),
        self._get_test_instruction(_QTransf.ADD_DEQUANTIZE),
        self._get_test_instruction(_QTransf.DUPLICATE_BUFFER, consumers=[2]),
        self._get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[3, 4]),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE),
        self._get_test_instruction(_QTransf.DUPLICATE_BUFFER, consumers=[3, 4]),
    ]
    tensor_instructions = qtyping.TensorTransformationInsts(
        tensor_name="test_tensor",
        subgraph_id=0,
        instructions=instructions,
    )
    instruction_gen = (
        instruction_generator.TransformationInstructionsGenerator()
    )
    instruction_gen._remove_unnecessary_buffer_duplication(tensor_instructions)

    self.assertLen(tensor_instructions.instructions, 6)
    expected_transformations = [
        _QTransf.DUPLICATE_TENSOR,
        _QTransf.ADD_QUANTIZE,
        _QTransf.ADD_DEQUANTIZE,
        _QTransf.DUPLICATE_BUFFER,
        _QTransf.DUPLICATE_TENSOR,
        _QTransf.ADD_QUANTIZE,
    ]
    got_transformations = [
        instruction.transformation
        for instruction in tensor_instructions.instructions
    ]
    self.assertEqual(got_transformations, expected_transformations)

  def test__instruction_generator_removes_unnecessary_tensor_and_buffer_duplication(
      self,
  ):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/constant_tensor_and_buffer_only_sharing_weight_fcs.tflite",
    )
    params_4_bits = qtyping.UniformQuantParams(
        4, None, np.array([1]), np.array([0])
    )
    params_8_bits = qtyping.UniformQuantParams(
        8, None, np.array([1]), np.array([0])
    )
    quant_parameters = {}
    # Two FCs share a weight tensor `arith.constant`.
    quant_parameters["arith.constant"] = qtyping.TensorTransformationParams(
        tensor_name="arith.constant",
        producer=None,
        consumers=[
            qtyping.OpToTensorParams(
                subgraph_op_id=0,
                transformations=[
                    _QTransf.DUPLICATE_TENSOR,
                    _QTransf.DUPLICATE_BUFFER,  # Expected to be removed.
                    _QTransf.QUANTIZE_TENSOR,
                ],
                parameters=params_8_bits,
            ),
            qtyping.OpToTensorParams(
                subgraph_op_id=1,
                transformations=[
                    _QTransf.DUPLICATE_TENSOR,  # Expected to be removed.
                    _QTransf.DUPLICATE_BUFFER,
                    _QTransf.QUANTIZE_TENSOR,
                ],
                parameters=params_4_bits,
            ),
        ],
    )
    instruction_gen = instruction_generator.TransformationInstructionsGenerator(
        test_model_path
    )
    instructions = instruction_gen.quant_params_to_transformation_insts(
        quant_parameters
    )

    def get_expected_instruction(transformation, consumers, params):
      return qtyping.TransformationInst(
          transformation=transformation,
          consumers=consumers,
          tensor_id=1,
          producer=-1,
          parameters=params,
      )

    expected_instructions = qtyping.TensorTransformationInsts(
        tensor_name="arith.constant",
        subgraph_id=0,
        instructions=[
            get_expected_instruction(
                _QTransf.DUPLICATE_TENSOR, consumers=[0], params=params_8_bits
            ),
            get_expected_instruction(
                _QTransf.DUPLICATE_BUFFER, consumers=[1], params=params_4_bits
            ),
            get_expected_instruction(
                _QTransf.QUANTIZE_TENSOR, consumers=[0], params=params_8_bits
            ),
            get_expected_instruction(
                _QTransf.QUANTIZE_TENSOR, consumers=[1], params=params_4_bits
            ),
        ],
    )
    self.assertLen(instructions, 1)
    self.assertEqual(instructions["arith.constant"], expected_instructions)

  def test__split_instructions_by_tensor_duplication_returns_expected_subsets(
      self,
  ):
    instructions = [
        self._get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[1, 2, 3]),  # pylint: disable=line-too-long
        self._get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[4]),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[1, 2]),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[3]),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[4]),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[5]),
    ]
    tensor_instructions = qtyping.TensorTransformationInsts(
        tensor_name="test_tensor", subgraph_id=0, instructions=instructions
    )
    instruction_gen = (
        instruction_generator.TransformationInstructionsGenerator()
    )
    got = instruction_gen._split_instructions_by_tensor_duplication(
        tensor_instructions
    )
    expected = [
        [self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[5])],
        [
            self._get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[1, 2, 3]),  # pylint: disable=line-too-long
            self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[1, 2]),
            self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[3]),
        ],
        [
            self._get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[4]),  # pylint: disable=line-too-long
            self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[4]),
        ],
    ]
    self.assertEqual(got, expected)

  def test__check_tensor_transformation_instructions_valid_succeeds_on_q_dq_with_duplication(
      self,
  ):
    instructions = [
        self._get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[1]),
        self._get_test_instruction(_QTransf.NO_QUANTIZE, consumers=[1]),
        self._get_test_instruction(_QTransf.QUANTIZE_TENSOR, consumers=[2]),
    ]
    tensor_instructions = qtyping.TensorTransformationInsts(
        tensor_name="test_tensor", subgraph_id=0, instructions=instructions
    )
    instruction_gen = (
        instruction_generator.TransformationInstructionsGenerator()
    )
    instruction_gen._check_tensor_transformation_instructions_valid(
        tensor_instructions
    )

  def test__check_tensor_transformation_instructions_valid_fails_when_q_noq_wo_duplication(
      self,
  ):
    tensor_instructions = qtyping.TensorTransformationInsts(
        tensor_name="test_tensor",
        subgraph_id=0,
        instructions=[
            self._get_test_instruction(_QTransf.NO_QUANTIZE, consumers=[1]),
            self._get_test_instruction(_QTransf.QUANTIZE_TENSOR, consumers=[2]),
        ],
    )
    instruction_gen = (
        instruction_generator.TransformationInstructionsGenerator()
    )
    with self.assertRaisesRegex(
        ValueError, "can not be both quantized and unquantized"
    ):
      instruction_gen._check_tensor_transformation_instructions_valid(
          tensor_instructions
      )


class EliminateUnnecessaryRequantizationTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.ins_gen = instruction_generator.TransformationInstructionsGenerator(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/conv_fc_mnist.tflite"
        )
    )

  def _get_test_instruction(
      self,
      transformation: qtyping.QuantTransformation,
      producer: int = -1,
      consumers: Optional[Sequence[int]] = None,
      qparams: Optional[qtyping.UniformQuantParams] = None,
  ) -> qtyping.TransformationInst:
    if consumers is None:
      consumers = []
    if qparams is None:
      qparams = qtyping.UniformQuantParams(
          num_bits=8,
          quantized_dimension=None,
          scale=np.array([1]),
          zero_point=np.array([0]),
      )
    return qtyping.TransformationInst(
        transformation=transformation,
        producer=producer,
        consumers=consumers,
        parameters=qparams,
        # Dummy values below.
        tensor_id=0,
    )

  def _create_test_insts(
      self, instructions: list[qtyping.TransformationInst]
  ) -> qtyping.TensorTransformationInsts:
    return qtyping.TensorTransformationInsts(
        tensor_name="test_tensor", subgraph_id=0, instructions=instructions
    )

  def test_no_fusion_when_too_few_instructions(self):
    tensor_insts = self._create_test_insts([
        self._get_test_instruction(_QTransf.QUANTIZE_TENSOR),
    ])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )
    self.assertLen(tensor_insts.instructions, 1)

  def test_no_fusion_when_too_many_instructions(self):
    tensor_insts = self._create_test_insts([
        self._get_test_instruction(_QTransf.QUANTIZE_TENSOR),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE),
        self._get_test_instruction(_QTransf.ADD_DEQUANTIZE),
    ])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )
    self.assertLen(tensor_insts.instructions, 3)

  def test_no_fusion_when_invalid_transformation_pair(self):
    tensor_insts = self._create_test_insts([
        self._get_test_instruction(_QTransf.ADD_DEQUANTIZE),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE),
    ])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )
    self.assertLen(tensor_insts.instructions, 2)

  def test_no_fusion_when_consumers_mismatch(self):
    tensor_insts = self._create_test_insts([
        self._get_test_instruction(_QTransf.QUANTIZE_TENSOR, consumers=[0]),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[1]),
    ])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )
    self.assertLen(tensor_insts.instructions, 2)

  def test_no_fusion_when_no_producer(self):
    producer = -1
    tensor_insts = self._create_test_insts([
        self._get_test_instruction(_QTransf.QUANTIZE_TENSOR, producer),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE, producer),
    ])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )
    self.assertLen(tensor_insts.instructions, 2)

  def test_no_fusion_when_quant_params_are_incompatible(self):
    params_8_bits = qtyping.UniformQuantParams(
        8, None, np.array([1]), np.array([0])
    )
    params_16_bits = qtyping.UniformQuantParams(
        16, None, np.array([1]), np.array([0])
    )
    tensor_insts = self._create_test_insts([
        self._get_test_instruction(
            _QTransf.QUANTIZE_TENSOR, qparams=params_8_bits
        ),
        self._get_test_instruction(
            _QTransf.ADD_QUANTIZE, qparams=params_16_bits
        ),
    ])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )
    self.assertLen(tensor_insts.instructions, 2)

  def test_no_fusion_when_producer_constrained(self):
    # Reshape op (op index 2) has same as input scale constraint.
    tensor_insts = self._create_test_insts([
        self._get_test_instruction(_QTransf.QUANTIZE_TENSOR, producer=2),
        self._get_test_instruction(_QTransf.ADD_QUANTIZE, producer=2),
    ])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )
    self.assertLen(tensor_insts.instructions, 2)

  def test_fusion_succeeds(self):
    producer = 0
    consumers = [1]
    params_0 = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=None,
        scale=np.array([1]),
        zero_point=np.array([0]),
    )
    params_1 = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=None,
        scale=np.array([2]),
        zero_point=np.array([1]),
    )
    inst_0 = self._get_test_instruction(
        _QTransf.QUANTIZE_TENSOR, producer, consumers, params_0
    )
    inst_1 = self._get_test_instruction(
        _QTransf.ADD_QUANTIZE, producer, consumers, params_1
    )
    tensor_insts = self._create_test_insts([inst_0, inst_1])
    self.ins_gen._eliminate_requantization_for_nonconstrained_provider(
        tensor_insts
    )

    self.assertLen(tensor_insts.instructions, 1)
    result_inst = tensor_insts.instructions[0]
    self.assertEqual(result_inst.transformation, _QTransf.QUANTIZE_TENSOR)

    result_params = result_inst.parameters
    # Explicitly narrow the type for pytype.
    if not isinstance(result_params, qtyping.UniformQuantParams):
      self.fail("Fused instruction parameters are not UniformQuantParams")

    self.assertEqual(result_params.scale, params_1.scale)
    self.assertEqual(result_params.zero_point, params_1.zero_point)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with div op."""

import pathlib

from absl.testing import parameterized
import absl.testing.absltest as absltest

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils


_TEST_MODEL_FOLDER = test_utils.get_path_to_datafile('../models/')
_QuantAlgo = quantizer.AlgorithmName


class DivTest(test_utils.BaseOpTestCase):

  def setUp(self):
    super().setUp()
    self._op_name = qtyping.TFLOperationName.DIV

  @parameterized.parameters(
      # algorithm_key, activations_num_bits, symmetric
      (_QuantAlgo.MIN_MAX_UNIFORM_QUANT, 8, True),
      (_QuantAlgo.MIN_MAX_UNIFORM_QUANT, 8, False),
      (_QuantAlgo.MIN_MAX_UNIFORM_QUANT, 16, True),
      (_QuantAlgo.OCTAV, 8, True),
      (_QuantAlgo.OCTAV, 16, True),
  )
  def test_static_quantization_accuracy_and_size_within_tolerance(
      self, algorithm_key, activations_num_bits, symmetric
  ):
    output_tolerance = 1e-3
    model_filename = 'single_div.tflite'
    model_path = pathlib.Path(_TEST_MODEL_FOLDER) / model_filename

    activation_config = test_utils.get_static_activation_quant_setting(
        activations_num_bits, symmetric
    )
    op_config = test_utils.get_static_op_quant_config(activation_config)
    self.assert_quantization_accuracy(
        algorithm_key=algorithm_key,
        model_path=model_path,
        op_name=self._op_name,
        op_config=op_config,
        output_tolerance=output_tolerance,
        num_validation_samples=1,
        num_calibration_samples=1,
    )


if __name__ == '__main__':
  absltest.main()
